#include <algorithm>
#include <iostream>
#include <string>
#include <vector>
#include <cstring>
#include <cstdio>
#include <ctime>
#include <cmath>

#include <FL/Fl.H>                      // Main FLTK header file
#include <FL/Fl_Group.H>                // Fl_Group header file
#include <FL/Fl_Window.H>               // Fl_Window header file
#include <FL/fl_ask.H>                  // FLTK convenience functions
#include <FL/fl_file_chooser.H>         // FLTK file chooser
#include <FL/Fl_Menu_Bar.H>       
#include <FL/Fl_Choice.H>
// for Export
#include <FL/Fl_Input.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Return_Button.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Gl_Window.H>
#include <Fl/math.h>

#include "vector3d.h"
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "pygmalion.h"
#include "glWindow.h"
#include "glWindowUI.h"
#include "camera.h"
#include "light.h"
#include "cameraset.h"
#include "mainmenu.h"
#include "newFile.h"
#include "usersetting.h"


#ifdef __MINGW32__
#define isnan _isnan
#endif
#define DebugOut(a); cout << a << endl;
using namespace std;

/////////////////////////////////////////////
// menu
CameraSet* cameraSet=NULL;
MainMenu* mainM=NULL;
vector<Fl_Menu_Item> menu;
UserSetting* userSetting=NULL;

Fl_Menu_Item* undoMA; // undo menu address
Fl_Menu_Item* redoMA; // redo menu address
Fl_Menu_Item* restX; // restrict X
Fl_Menu_Item* restY;
Fl_Menu_Item* restZ;


//////////////////////////////////////////////
// needed for file operation
char defaultSub[255]; // default subdivide level
int changed = 0;
char filename[1024] = ""; // needed in function Load(...)
char expFilename[1024] = ""; // export file name or temporary
//char tmpExpFilename[1024] = ""; // for direct render from pygmalion
char renderExe[512]; // user specify renderer
//char renderOpt[512]; // user specify render option
char typeOpt[512]; // auto specific render option

int vertexOffset;  
int edgeOffset;
int faceOffset;

///////////////////////////////////////////////////
// UNDO
vector<int> stackOP; // stack of operation for UNDOing
vector<vector<Vertex> > undoVertex; // undo buffer
vector<vector<Edge> > undoEdge;
vector<vector<Face> > undoFace;
vector<vector<int> > undoSelectV;
vector<vector<int> > undoSelectE;
vector<vector<int> > undoSelectF;

enum {
  uGEOMETRY=1,
  uSELECTION=2,
  uCAMERA=4,
  uTEXTURE=8
}opCode;

int lastGEOOP=-1; // index in last operation(lastOP reserved when Undo)
int lastSELOP=-1; // -1 means no operation
int lastCAMOP=-1;
int lastTEXOP=-1;
int lastOPIndex=-1; // specify last index of operation stack

////////////////////////////////////////////////////
// geometry
vector<Vertex> vertex; // Vertex Vector
vector<Edge> edge; // Edge Vector
vector<Face> face; // Face Vector

// order of Vertex in subdivide level
// ex.{10,20,45} -> Level 1 {0-9}, Level 2 {10-19}, Level 3 {20-44}
vector<int> subdividedVertex; 
vector<int> subdividedEdge; // Edge
vector<int> subdividedFace; // Face
int subdivideLevel=0; // init value=0 mean no subdivision
int nowLevel; // preserve subdivideLevel

///////////////////////////////////////////////////
// selection
// index of selected object
vector<int> selectV;
vector<int> selectE;
vector<int> selectF;
vector<int> totalV; // total vertex within select{V,E,F} for TransMode
extern enum {
  sVertex,sEdge,sFace
}sMode; // select target

/////////////////////////////////////////////////
// mode of left button of mouse
const int sNORMAL=0;
const int sSUBDIVIDE=1;
int nowStatus=sNORMAL;

extern vector<Camera> cameraList;
int outputC=0; // output camera
extern vector<Light> lightList;

vector<GLWindowUI*> glW;

// bounding box for clipping for drawing by OpenGL
extern double minX,minY,minZ;
extern double maxX,maxY,maxZ;
// bounding box of selected object ; now only vertex
double minXS,maxXS,minYS,maxYS,minZS,maxZS; // {min,max}{X,Y,Z} of selection


/////////////////////////////////////////////////////
void
Render(){
  char renderScript[1024];
  if(!strcmp(renderExe,"")){
    fl_alert("Renderer not set! Please set renderer.");
    userSetting->mainWin->show();
    userSetting->renderTab->show();
    userSetting->mainWin->set_modal();
    return;
  }
  if(cameraList.size()==0){
    fl_alert("Camera not found... Please create camera.");
    return;
  }

  strcpy(renderScript,renderExe); // execute renderer
  strcat(renderScript," ");
  strcat(renderScript,userSetting->renderOptI->value()); // render option
  strcat(renderScript," ");
  strcat(renderScript,typeOpt); // renderer type option
  strcat(renderScript," ");
  char tmpWidth[10];
  char tmpHeight[10];
  sprintf(tmpWidth,"+W%d ",cameraList[outputC].width);
  sprintf(tmpHeight,"+H%d ",cameraList[outputC].height);  
  strcat(renderScript,tmpWidth);
  strcat(renderScript,tmpHeight);
  if(userSetting->pauseCheck->value()){
    strcat(renderScript,"+P ");
  }
  strcat(renderScript,"+D +I");
  strcat(renderScript,tmpExpFilename);
  strcat(renderScript," &");

  ExportPOVCore(atoi(userSetting->subTimesI->value()),TEMP_EXPORT);

  system(renderScript);
}


// vector copy -- Vertex,Edge,Face
template<typename T>
void VectorCopy(vector<T>& src,vector<T>& des,int begin,int end){
  des.erase(des.begin(),des.end());
  for(int cnt1=begin;cnt1<end;cnt1++){
    des.push_back(src[cnt1]);
  }
}

void
Unique(vector<int>& vec){
  for(int cnt1=0;cnt1<vec.size();cnt1++){
    int removeNum=vec[cnt1];
    for(int cnt2=cnt1+1;cnt2<vec.size();cnt2++){
      if(vec[cnt2]==removeNum){
	vec.erase(&vec[cnt2]);
	cnt2--;
      }
    }
  }
}

void
RefineTotalV(){
  totalV.erase(totalV.begin(),totalV.end());
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    totalV.push_back(selectV[cnt1]);
  }
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    for(int cnt2=0;cnt2<edge[selectE[cnt1]].vertexLink.size();cnt2++){
      totalV.push_back(edge[selectE[cnt1]].vertexLink[cnt2]);
    }
  }
  for(int cnt1=0;cnt1<selectF.size();cnt1++){
    for(int cnt2=0;cnt2<face[selectF[cnt1]].vertexLink.size();cnt2++){
      totalV.push_back(face[selectF[cnt1]].vertexLink[cnt2]);
    }
  }
  
  // My Unique Algorithm
  Unique(totalV);
  //cout << "totalV" << endl;
  //PrintVec(totalV);
}

// called after geometry change,selection,,,
void
PushOP(int op){
  DebugOut("PushOP : lastOPIndex==" << lastOPIndex);
  stackOP.erase(&stackOP[lastOPIndex+1],stackOP.end());
  if(op&uGEOMETRY){
    DebugOut("geometry : lastGEOOP==" << lastGEOOP);
    undoVertex.erase(&undoVertex[lastGEOOP+1],undoVertex.end());
    undoEdge.erase(&undoEdge[lastGEOOP+1],undoEdge.end());
    undoFace.erase(&undoFace[lastGEOOP+1],undoFace.end());
    undoVertex.push_back();
    VectorCopy<Vertex>(vertex,undoVertex.back(),0,subdividedVertex[1]);
    undoEdge.push_back();
    VectorCopy<Edge>(edge,undoEdge.back(),0,subdividedEdge[1]);
    undoFace.push_back();
    VectorCopy<Face>(face,undoFace.back(),0,subdividedFace[1]);
    lastGEOOP++;
  }
  if(op&uSELECTION){
    DebugOut("selection : lastSELOP==" << lastSELOP);
    undoSelectV.erase(&undoSelectV[lastSELOP+1],undoSelectV.end());
    undoSelectE.erase(&undoSelectE[lastSELOP+1],undoSelectE.end());
    undoSelectF.erase(&undoSelectF[lastSELOP+1],undoSelectF.end());
    undoSelectV.push_back();
    VectorCopy<int>(selectV,undoSelectV.back(),0,selectV.size());
    undoSelectE.push_back();
    VectorCopy<int>(selectE,undoSelectE.back(),0,selectE.size());
    undoSelectF.push_back();
    VectorCopy<int>(selectF,undoSelectF.back(),0,selectF.size());
    lastSELOP++;
  }
  if(op&uCAMERA){   
    //cout << "camera" << endl;
  }   
  if(op&uTEXTURE){
    //cout << "texture" << endl;
  }
  stackOP.push_back(op);
  lastOPIndex++;
  if(stackOP.size()>1){
    undoMA->activate();
  }
  redoMA->deactivate();

  mainM->mainWin->redraw();
  //cout << "PushOP after : lastOPIndex==" << lastOPIndex << endl;
  //cout << "stackOP.size()==" << stackOP.size() << endl;

  /*
  for(int cnt1=0;cnt1<undoVertex.size();cnt1++){
    for(int cnt2=0;cnt2<undoVertex[cnt1].size();cnt2++){
      undoVertex[cnt1][cnt2].Print();
    }
    cout << endl;
  }
  cout << endl;
  */
}

// don't call with lastOPIndex<=-1
// so this function should be called by mainM->undoM
void
Undo(){
  //cout << "Undo : lastOPIndex==" << lastOPIndex << endl;
  if((stackOP[lastOPIndex]&uGEOMETRY)){
    //cout << "geometry : lastGEOOP==" << lastGEOOP << endl;
    PreSub();
    VectorCopy<Vertex>(undoVertex[lastGEOOP-1],vertex,0,undoVertex[lastGEOOP-1].size());
    VectorCopy<Edge>(undoEdge[lastGEOOP-1],edge,0,undoEdge[lastGEOOP-1].size());
    VectorCopy<Face>(undoFace[lastGEOOP-1],face,0,undoFace[lastGEOOP-1].size());
    subdividedVertex[1]=vertex.size();
    subdividedEdge[1]=edge.size();
    subdividedFace[1]=face.size();
    PostSub();
    lastGEOOP--;
  }
  if((stackOP[lastOPIndex]&uSELECTION)){
    //cout << "selection : lastSELOP==" << lastSELOP << endl;
    VectorCopy<int>(undoSelectV[lastSELOP-1],selectV,0,undoSelectV[lastSELOP-1].size());    
    VectorCopy<int>(undoSelectF[lastSELOP-1],selectF,0,undoSelectF[lastSELOP-1].size());    
    VectorCopy<int>(undoSelectE[lastSELOP-1],selectE,0,undoSelectE[lastSELOP-1].size());    
    RefineTotalV();
    lastSELOP--;
  }
  if((stackOP[lastOPIndex]&uCAMERA)){
  }
  if((stackOP[lastOPIndex]&uTEXTURE)){
  }
  lastOPIndex--;
  //cout << "Undo after : lastOPIndex==" << lastOPIndex << endl;
  //cout << "stackOP.size()==" << stackOP.size() << endl;
  redoMA->activate();
  if(lastOPIndex==0){ // nothing to undo
    undoMA->deactivate();
  }
  /*
  for(int cnt1=0;cnt1<undoVertex.size();cnt1++){
    for(int cnt2=0;cnt2<undoVertex[cnt1].size();cnt2++){
      undoVertex[cnt1][cnt2].Print();
    }
    cout << endl;
  }
  cout << endl;
  */
  Sync();
}

void
Redo(){
  //cout << "Redo : lastOPIndex==" << lastOPIndex << endl;
  if((stackOP[lastOPIndex+1]&uGEOMETRY)){
    //cout << "geometry" << endl;
    PreSub();
    VectorCopy<Vertex>(undoVertex[lastGEOOP+1],vertex,0,undoVertex[lastGEOOP+1].size());
    VectorCopy<Edge>(undoEdge[lastGEOOP+1],edge,0,undoEdge[lastGEOOP+1].size());
    VectorCopy<Face>(undoFace[lastGEOOP+1],face,0,undoFace[lastGEOOP+1].size());
    subdividedVertex[1]=vertex.size();
    subdividedEdge[1]=edge.size();
    subdividedFace[1]=face.size();
    PostSub();
    lastGEOOP++;
  }
  if((stackOP[lastOPIndex+1]&uSELECTION)){
    //cout << "selection" << endl;
    VectorCopy<int>(undoSelectV[lastSELOP+1],selectV,0,undoSelectV[lastSELOP+1].size());
    VectorCopy<int>(undoSelectE[lastSELOP+1],selectE,0,undoSelectE[lastSELOP+1].size());
    VectorCopy<int>(undoSelectF[lastSELOP+1],selectF,0,undoSelectF[lastSELOP+1].size());
    RefineTotalV();
    lastSELOP++;
  }
  if((stackOP[lastOPIndex+1]&uCAMERA)){
  }
  if((stackOP[lastOPIndex+1]&uTEXTURE)){
  }
  lastOPIndex++;
  undoMA->activate();
  if(lastOPIndex==stackOP.size()-1){ // nothing to undo
    redoMA->deactivate();
  }
  //cout << "Redo after : lastOPIndex==" << lastOPIndex << endl;
  //cout << "stackOP.size()==" << stackOP.size() << endl;
  /*
  for(int cnt1=0;cnt1<undoVertex.size();cnt1++){
    for(int cnt2=0;cnt2<undoVertex[cnt1].size();cnt2++){
      undoVertex[cnt1][cnt2].Print();
    }
    cout << endl;
  }
  cout << endl;
  */
  Sync();
}

void
CheckPointer(){
    for(int cnt1=0;cnt1<glW.size();cnt1++){
/*
	cout << "glW[" << cnt1 << "]=" << glW[cnt1] << endl;
	cout << "glW[" << cnt1 << "]->glWindow=" << glW[cnt1]->glWindow << endl;
	cout << "glW[" << cnt1 << "]->glWindow->camera=" << glW[cnt1]->glWindow->camera << endl;
	cout << "cameraList[" << cnt1 << "]=" << &cameraList[cnt1] << endl;
*/
	if(&cameraList[cnt1]!=glW[cnt1]->glWindow->camera){
	  fl_alert("cameraList[cnt1]!=glW[cnt1]->glWindow->camera !!!");
	}
	for(int cnt2=cnt1+1;cnt2<glW.size();cnt2++){
	  if(glW[cnt1]->glWindow==glW[cnt2]->glWindow){
	    fl_alert("glW[%d] and glW[%d] share same glWindow!!!",cnt1,cnt2);
	  }
	}
	//glW[cnt1]->glWindow
    }
}


void
CheckBBSCore(Vector3d& v){
  // X
  if(v.X()>maxXS){
    maxXS=v.X();
  }
  if(v.X()<minXS){
    minXS=v.X();
  }
  // Y
  if(v.Y()>maxYS){
    maxYS=v.Y();
  }
  if(v.Y()<minYS){
    minYS=v.Y();
  }
  // Z
  if(v.Z()>maxZS){
    maxZS=v.Z();
  }
  if(v.Z()<minZS){
    minZS=v.Z();
  }
}

void
CheckBBS(){
  minXS=DBL_MAX/2;maxXS=-1*DBL_MAX/2;
  minYS=DBL_MAX/2;maxYS=-1*DBL_MAX/2;
  minZS=DBL_MAX/2;maxZS=-1*DBL_MAX/2;
  for(int cnt1=0;cnt1<totalV.size();cnt1++){
    CheckBBSCore(vertex[totalV[cnt1]]);
  }
}

void
CenteringObject(){
  CheckBBS();
  Vector3d center;
  center.Set((minXS+maxXS)/2,(minYS+maxYS)/2,(minZS+maxZS)/2);    
  for(int cnt1=0;cnt1<totalV.size();cnt1++){
    vertex[totalV[cnt1]].Minus(center);
  }
  Sync();
}

void
AlignViews(){
  int startX,startY;
  int topWin=0;
  startX=mainM->mainWin->x();
  startY=mainM->mainWin->y()+mainM->mainWin->h()+glOffsetB+glOffsetT;
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    if(!glW[cnt1]->mainWindow->visible()){
      continue;
    }
    glW[cnt1]->mainWindow->position(startX,startY);
    startY+=glW[cnt1]->mainWindow->h()+glOffsetT+glOffsetB;
    if(cnt1<glW.size()-1){
      //if(startY+glW[cnt1+1]->mainWindow->h()+glOffsetT+glOffsetB>Fl::h()){
      if(startY+glW[cnt1+1]->mainWindow->h()+glOffsetB>Fl::h()){
	startX+=glW[topWin]->mainWindow->w()+glOffsetR+glOffsetL;
	startY=mainM->mainWin->y()+mainM->mainWin->h()+glOffsetB+glOffsetT;
	topWin=cnt1+1;
      }
    }
  }
}

// POV-Ray's angle means window's width angle,
// but OpenGL's angle means window's height angle.
double
POVAng2OGLAng(Camera& cam,double povAng){
  double oglAng;
  oglAng=atan((double)cam.height/cam.width*tan(povAng/2*M_PI/180))
    /M_PI*180*2.0;
  return oglAng;
}

double
OGLAng2POVAng(Camera& cam,double oglAng){
  double povAng;
  povAng=atan((double)cam.width/cam.height*tan(oglAng/2*M_PI/180))
    /M_PI*180*2.0;
  return povAng;
}

// should be called after creating camera
void
CreateGLWUI(){ 
  // menu modification1
  mainM->AddViewIndex();
  // GLWindouUI initialize
  int cnt=(int)cameraList.size()-1; // >=0
  // retain glW->visible()
  vector<int> viewFlag;
  for(int cnt1=0;cnt1<viewFlag.size();cnt1++){
    viewFlag.push_back(glW[cnt1]->mainWindow->visible());
    glW[cnt1]->glWindow->hide();
  }
  glW.push_back();
  glW[cnt] = new GLWindowUI;
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    glW[cnt1]->SetCamera(cnt1,cameraList[cnt1].width,cameraList[cnt1].height);
  }
  glW[cnt]->hide();
  glW[cnt]->show();
  for(int cnt1=0;cnt1<viewFlag.size();cnt1++){
    if(viewFlag[cnt1]){
      glW[cnt1]->show();
    }
  }
  // menu modification2
  mainM->RenewViewIndex();
}

void
CreateView(int type){
  switch(type){
  case -1:
    break;
  default:
    CreateCamera(type);
    CreateGLWUI();
    break;
  }
  CheckPointer();
}

void
PreDeleteView(Fl_Menu_Item*,int num){
  vector<int> viewFlag;
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    viewFlag.push_back(glW[cnt1]->mainWindow->visible());
    glW[cnt1]->hide();
  }
  glW[num]->show();
  if(!fl_ask("Delete this View...")){
    for(int cnt1=0;cnt1<glW.size();cnt1++){
      if(viewFlag[cnt1]){
	glW[cnt1]->show();
      }else{
	glW[cnt1]->hide();
      }
    }
    return;
  }
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    if(viewFlag[cnt1]){
      glW[cnt1]->show();
    }else{
      glW[cnt1]->hide();
    }
  }
  DeleteView(num);
}

void
DeleteView(int num){
  // cameraList
  cameraList.erase(&cameraList[num]);
  // glW
  glW[num]->hide();
  delete glW[num];
  glW.erase(&glW[num]);
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    glW[cnt1]->SetCamera(cnt1,cameraList[cnt1].width,cameraList[cnt1].height);
  }
  // view index
  mainM->DeleteViewIndex(num);
  mainM->RenewViewIndex();
}

void
ToggleView(Fl_Menu_Item*,int num){
  if(glW[num]->mainWindow->visible()){
    glW[num]->hide();
  }else{
    glW[num]->show();
  }
}

void
CreateCamera(int type){
  switch(type){
  case PERSPECTIVE:
    cameraList.push_back(Camera());
    cameraList.back().camera.Set(20,10,90); // Warning! z is inverse 
    cameraList.back().view.Set(0,0,0);
    break;
  case XYPM:
  case XYMP:
    cameraList.push_back(Camera());
    cameraList.back().type=type;
    cameraList.back().camera.Set(0,0,10); // Warning! z is inverse
    break;
  case YZPM:
  case YZMP:
    cameraList.push_back(Camera());
    cameraList.back().type=type;
    cameraList.back().camera.Set(10,0,0); // Warning! z is inverse 
    break;
  case XZPM:
  case XZMP:
    cameraList.push_back(Camera());
    cameraList.back().type=type;
    cameraList.back().camera.Set(0,10,0); // Warning! z is inverse   
    break;
  }
}

void
CreateNewWin(){
   cameraSet->ReadValue(0);  // 0 mean default value
   cameraSet->cameraSetW->set_modal();
   cameraSet->show();  
}

void
ChangeWinProp(){
#ifdef DEBUG
  cout << "ChangeWinProp" << endl;
#endif
  cameraSet->ReadValue(1);  // 1 mean using camera[outputC]
  cameraSet->cameraSetW->set_modal();
  cameraSet->show();  
}

// Add Edge and Connect Vertex(vNum1_) between vertex(vNum2_)
void
AddEdge(int vNum1_,int vNum2_){
#ifdef DEBUG
  if(0 > vNum1_>=vertex.size() || 0 > vNum2_>=vertex.size()){
    cout << "strange!!!!!!!!!!!! in AddEdge." << endl;
  }
#endif
  if(vNum1_==vNum2_){
#ifdef DEBUG
    cout << "same vertex can't be connected." << endl;
#endif
    return;
  }
  // This makes " vNum1 < vNum2 "
  if(vNum1_>vNum2_){
    swap(vNum1_,vNum2_);
  }

  if(nowStatus==sNORMAL){
    // Find pre generated edge.
    int flag=FindEdge((const int)vNum1_,(const int)vNum2_);
    if(flag!=-1){
#ifdef DEBUG
      cout << "These vertices already connected!!!!!!!!!!!!!!!!!!!" << endl;
      cout << "vNum1,vNum2==" << vNum1_ << "," << vNum2_ << endl;
#endif
      return;
    }
  }
  edge.push_back(Edge(vNum1_,vNum2_));
  subdividedEdge[subdivideLevel+1]++;
  SetChanged(1);
  vertex[vNum1_].SetEdgeLink();
  vertex[vNum2_].SetEdgeLink();

  // 
  int flag2;
  flag2=edge[edge.size()-1].FindAddFace(); // {three,four} vertices
  // if it subdivides four Edges Face,remove the face.
  if(flag2){
    RemoveExtraFace();
  }
}

void
RemoveExtraFace(){
  int v0,v1;

  for(int cnt1=0;cnt1<face.size();cnt1++){
    if(face[cnt1].edgeLink.size()==4){
      // first vertex pair
#ifdef DEBUG
      cout << "first vertex pair." << endl;
#endif
      v0=face[cnt1].vertexLink[0];
      v1=face[cnt1].vertexLink[2];
      if(v0>v1){
	swap(v0,v1);
      }
      if(FindEdge(v0,v1)!=-1){
#ifdef DEBUG
	cout << "Found three Edges Face within four Edges Face." << endl;
#endif
	RemoveFace(cnt1);
	cnt1-=1;
#ifdef DEBUG
	cout << "Four edge Face Erased." << endl;
#endif
      }else {
	// second vertex
#ifdef DEBUG
	cout << "second vertex pair." << endl;
#endif
	v0=face[cnt1].vertexLink[1];
	v1=face[cnt1].vertexLink[3];
	if(v0>v1){
	  swap(v0,v1);
	}
	if(FindEdge(v0,v1)!=-1){
#ifdef DEBUG
	  cout << "Found three Edges Face within four Edges Face." << endl;
#endif
	  RemoveFace(cnt1);
	  cnt1-=1;
#ifdef DEBUG
	  cout << "Four Edges Face Erased." << endl;
#endif
	}
      }
    }
  }
}

void
RemoveSelected(){
  // Subdivide Level -> 0
  for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
    Subdivide(-1);
  }

  // Face Remove Operation /////////////////////////////////
  // Clearing removeFace
  //removeFace.erase(removeFace.begin(),removeFace.end());
  vector<int> removeFace; // be removed by "delete"
  
  // Making Remove List of Face
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    for(int cnt2=0;cnt2<edge[selectE[cnt1]].facePLink.size();cnt2++){
      removeFace.push_back(edge[selectE[cnt1]].facePLink[cnt2]);
    }
  }
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[selectV[cnt1]].facePLink.size();cnt2++){
      removeFace.push_back(vertex[selectV[cnt1]].facePLink[cnt2]);
    }
  }
#ifdef DEBUG
  cout << "Fat removeFace List." << endl;
  //PrintVec(removeFace);
#endif
  // Erasing extra number
  Unique(removeFace);
  /*
  int removeNum;
  for(int cnt1=0;cnt1<removeFace.size();cnt1++){
    removeNum=removeFace[cnt1];
    for(int cnt2=cnt1+1;cnt2<removeFace.size();cnt2++){
      if(removeFace[cnt2]==removeNum){
	removeFace.erase(&removeFace[cnt2]);
	cnt2--;
      }
    }
  }
  */
#ifdef DEBUG
  cout << "Smart removeFace List." << endl;
  //PrintVec(removeFace);
#endif
  for(int cnt1=0;cnt1<removeFace.size();cnt1++){
    RemoveFace(removeFace[cnt1]);
    for(int cnt2=cnt1;cnt2<removeFace.size();cnt2++){
      if(removeFace[cnt2]>removeFace[cnt1]){
	removeFace[cnt2]-=1;
      }
    }
  }

  // Edge Remove Operation ///////////////////////////////////////
  // Adding Remove List of Edge
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[selectV[cnt1]].edgePLink.size();cnt2++){
      selectE.push_back(vertex[selectV[cnt1]].edgePLink[cnt2]);
    }
  }
#ifdef DEBUG
  cout << "extra selectE" << endl;
#endif
  // Erasing extra number
  Unique(selectE);
  /*
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    removeNum=selectE[cnt1];
    for(int cnt2=cnt1+1;cnt2<selectE.size();cnt2++){
      if(selectE[cnt2]==removeNum){
	selectE.erase(&selectE[cnt2]);
	cnt2--;
      }
    }
    }
    */
#ifdef DEBUG
  cout << "smart selectE" << endl;
#endif
  // Erasing Edge entity
  int lV,rV; // Right Vertex,Left Vertex
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    lV=edge[selectE[cnt1]].vertexLink[Left];
    rV=edge[selectE[cnt1]].vertexLink[Right];
    RemoveEdge(selectE[cnt1]);
    for(int cnt2=cnt1;cnt2<selectE.size();cnt2++){ // mod
      if(selectE[cnt2]>selectE[cnt1]){
	selectE[cnt2]-=1;
      }
    }
    Find4Face(lV,rV,cnt1);
  }

  // Vertex Remove Operation /////////////////////////////////////////
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    RemoveVertex(selectV[cnt1]);
    for(int cnt2=cnt1;cnt2<selectV.size();cnt2++){
      if(selectV[cnt2]>selectV[cnt1]){
	selectV[cnt2]-=1;
      }
    }
  }

  selectF.erase(selectF.begin(),selectF.end());  
  selectE.erase(selectE.begin(),selectE.end());
  selectV.erase(selectV.begin(),selectV.end());
  DebugOut("Removed all selected object.");
}

// This Procedure is called only by RemoveExtraFace !
void
RemoveFace(int num){
#ifdef DEBUG
  if(0>num>=face.size()){
    cout << "Strange!!!!!!! in RemoveFace" << endl;
  }
#endif
  for(int cnt1=0;cnt1<face[num].edgeLink.size();cnt1++){
    edge[face[num].edgeLink[cnt1]].UnSetFaceLink(num);
  }
  for(int cnt1=0;cnt1<face[num].vertexLink.size();cnt1++){
    vertex[face[num].vertexLink[cnt1]].UnSetFaceLink(num);
  }

  face.erase(&(face[num]));
  subdividedFace[subdivideLevel+1]--;

  // Modify Face order in facePLink of {Edges,Vertices}
  // vertex
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[cnt1].facePLink.size();cnt2++){
      if(vertex[cnt1].facePLink[cnt2]>num){
	vertex[cnt1].facePLink[cnt2]-=1;
      }
    }
  }
  // edge
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    for(int cnt2=0;cnt2<edge[cnt1].facePLink.size();cnt2++){
      if(edge[cnt1].facePLink[cnt2]>num){
	edge[cnt1].facePLink[cnt2]-=1;
      }
    }
  }
}

// Calling this procedure,
// no Face has num in edgeLink
void
RemoveEdge(int num){
#ifdef DEBUG
  if(0>num>=edge.size()){
    cout << "Strange!!!!!!!!! in RemoveEdge " << endl;
  }
#endif
  for(int cnt1=0;cnt1<edge[num].vertexLink.size();cnt1++){
    vertex[edge[num].vertexLink[cnt1]].UnSetEdgeLink(num);
  }
  edge.erase(&(edge[num]));
  subdividedEdge[subdivideLevel+1]--;

  // Modify Edge order in edgeLink,edgePLink of {Faces,Vertices}
  // Face
  for(int cnt1=0;cnt1<face.size();cnt1++){
    for(int cnt2=0;cnt2<face[cnt1].edgeLink.size();cnt2++){
      if(face[cnt1].edgeLink[cnt2]>num){
	face[cnt1].edgeLink[cnt2]-=1;
      }
    }
  }
  // Vertex
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[cnt1].edgePLink.size();cnt2++){
      if(vertex[cnt1].edgePLink[cnt2]>num){
	vertex[cnt1].edgePLink[cnt2]-=1;
      }
    }
  }
}

void
RemoveVertex(int num){
#ifdef DEBUG
  if(0>num>=vertex.size()){
    cout << "Strange!!!!!!!!! in RemoveVertex" << endl;
  }
#endif
  vertex.erase(&(vertex[num]));
  subdividedVertex[subdivideLevel+1]--;

  // Modify Face order in facePLink of Edges
  // edge
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    for(int cnt2=0;cnt2<edge[cnt1].vertexLink.size();cnt2++){
      if(edge[cnt1].vertexLink[cnt2]>num){
	edge[cnt1].vertexLink[cnt2]-=1;
      }
    }
  }

  for(int cnt1=0;cnt1<face.size();cnt1++){
    for(int cnt2=0;cnt2<face[cnt1].vertexLink.size();cnt2++){
      if(face[cnt1].vertexLink[cnt2]>num){
	face[cnt1].vertexLink[cnt2]-=1;
      }
    }
  }
}

// condition : edge1 < edge2
//inline
int
FindEdge(const int v1,const int v2){
  int cnt1;
#ifdef DEBUG
  if(0>v1>=vertex.size()||0>v2>=vertex.size()){
    cout << "strange!!!!!!! in FindEdge." << endl;    
  }
#endif
  /*
    while(cnt1<edge_.size()){
    if(edge_[cnt1].vertexLink[Left] == edge1){
    if(edge_[cnt1].vertexLink[Right] == edge2){
    return cnt1;
    }
    }
    cnt1++;
    }
    */
  /*
    for(cnt1=0;cnt1<edge_.size();cnt1++){
    if(edge_[cnt1].vertexLink[Left] == edge1){
    if(edge_[cnt1].vertexLink[Right] == edge2){
    return cnt1;
    }
    }
    }
    */
  /*
    int endCnt=edge.size();
    for(cnt1=0;cnt1<endCnt;cnt1++){
    if(edge[cnt1].vertexLink[Left] == edge1){
    if(edge[cnt1].vertexLink[Right] == edge2){
    return cnt1;
    }
    }
    }
    */
  int endCnt=edge.size();
  for(cnt1=0;cnt1<endCnt;cnt1++){
    if(edge[cnt1].vertexLink[0] == v1){
      if(edge[cnt1].vertexLink[1] == v2){
	return cnt1;
      }
    }
  }

  return -1;
}

void
Sort3(int& eNum1_,int& eNum2_,int& eNum3_){
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
}

// This procedure should be called under ordered + or - eNum? 
void
AddFace(int eNum1_,int eNum2_,int eNum3_,int vNum1,int vNum2,int vNum3){
  // To do {eNum1_ < eNum2_ < eNum3_}
  Sort3(eNum1_,eNum2_,eNum3_);

  face.push_back(Face(eNum1_,eNum2_,eNum3_,vNum1,vNum2,vNum3));
  subdividedFace[subdivideLevel+1]++;
  SetChanged(1);

  // This make Connected Vertices known their Face.
  vertex[vNum1].SetFaceLink(); 
  vertex[vNum2].SetFaceLink(); 
  vertex[vNum3].SetFaceLink();   
  
  edge[eNum1_].SetFaceLink();
  edge[eNum2_].SetFaceLink();
  edge[eNum3_].SetFaceLink();
}


// Add Face and Connect Edge{eNum1_,eNum2_,eNum3_,eNum4}
void
AddFace(int eNum1_,int eNum2_,int eNum3_,int eNum4_,
	int vNum1,int vNum2,int vNum3,int vNum4){
  // To do {eNum1_ < eNum2_ < eNum3_ < eNum4_}
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum3_>eNum4_){
    swap(eNum3_,eNum4_);
  }
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum3_>eNum4_){
    swap(eNum3_,eNum4_);
  }
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum3_>eNum4_){
    swap(eNum3_,eNum4_);
  }
  //4 3 2 1 -> 3 4 2 1 -> 3 2 4 1 -> 3 2 1 4 ->
  //2 3 1 4 -> 2 1 3 4 -> 1 2 3 4 

  face.push_back(Face(eNum1_,eNum2_,eNum3_,eNum4_,vNum1,vNum2,vNum3,vNum4));
  subdividedFace[subdivideLevel+1]++;
  SetChanged(1);
  
  vertex[vNum1].SetFaceLink();
  vertex[vNum2].SetFaceLink();
  vertex[vNum3].SetFaceLink();
  vertex[vNum4].SetFaceLink();

  edge[eNum1_].SetFaceLink();
  edge[eNum2_].SetFaceLink();
  edge[eNum3_].SetFaceLink();
  edge[eNum4_].SetFaceLink();
}

void
Find4Face(int lV,int rV,int edgeNum){
  Vertex* v1=&vertex[lV];
  Vertex* v2=&vertex[rV];
  vector<int> connectEdge1;
  vector<int> connectEdge2;
  vector<int> connectV;
  int awayV1,awayV2;
  
  for(int cnt1=0;cnt1<v1->edgePLink.size();cnt1++){
    edge[v1->edgePLink[cnt1]];
    awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Left];
    if(awayV1==lV){
      awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Right];
    }
    for(int cnt2=0;cnt2<v2->edgePLink.size();cnt2++){
      edge[v2->edgePLink[cnt2]];
      awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Left];
      if(awayV2==rV){
	awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Right];
      }      
      if(awayV1==awayV2){ // found connected !
#ifdef DEBUG
	cout << "found connected! in Find4Face." << endl;
#endif
	connectEdge1.push_back(v1->edgePLink[cnt1]);
	connectEdge2.push_back(v2->edgePLink[cnt2]);
	connectV.push_back(awayV1);
      }
    }
  }
#ifdef DEBUG
  cout << "connectEdge1" << endl;
  PrintVec(connectEdge1);
  cout << "connectEdge2" << endl;
  PrintVec(connectEdge2);
  cout << "connectV" << endl;
  PrintVec(connectV);
  cout << "selectE" << endl;
  PrintVec(selectE);
#endif
  
  int flag;
  for(int cnt1=0;cnt1<connectV.size();cnt1++){
    for(int cnt2=cnt1+1;cnt2<connectV.size();cnt2++){
      flag=0;

      for(int cnt3=edgeNum+1;cnt3<selectE.size();cnt3++){
	if(connectEdge1[cnt1]==selectE[cnt3]){
	  flag=1;
	}
	if(connectEdge1[cnt2]==selectE[cnt3]){
	  flag=1;
	}
	if(connectEdge2[cnt1]==selectE[cnt3]){
	  flag=1;
	}
	if(connectEdge2[cnt2]==selectE[cnt3]){
	  flag=1;
	}
      }
      DebugOut("flag==" << flag);
      if(flag==1){
	continue;
      }
      AddFace(connectEdge1[cnt1],connectEdge1[cnt2],
	      connectEdge2[cnt1],connectEdge2[cnt2],
	      lV,connectV[cnt1],rV,connectV[cnt2]);
    }
  }
}

void
PrintHelp(){
  cout << "Usage:" << endl;
  cout << "'s'     toggle [subdivision surface] or [polygon]  ." << endl;
  cout << "" << endl;
  cout << "" << endl;
}

void
SelectAll(){
  switch(sMode){
  case sVertex:
    selectV.erase(selectV.begin(),selectV.end());
    for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
      selectV.push_back(cnt1);
    }
    break;
  case sEdge:
    selectE.erase(selectE.begin(),selectE.end());
    for(int cnt1=0;cnt1<subdividedEdge[1];cnt1++){
      selectE.push_back(cnt1);
    }
    break;
  case sFace:
    selectF.erase(selectF.begin(),selectF.end());
    for(int cnt1=0;cnt1<subdividedFace[1];cnt1++){
      selectF.push_back(cnt1);
    }
    break;
  }
  RefineTotalV();
}

void
DeselectAll(){
  switch(sMode){
  case sVertex:
    selectV.erase(selectV.begin(),selectV.end());
    break;
  case sEdge:      
    selectE.erase(selectE.begin(),selectE.end());
    break;
  case sFace:
    selectF.erase(selectF.begin(),selectF.end());      
    break;
  }
  RefineTotalV();
}

void
DeselectAllofAll(){
  selectV.erase(selectV.begin(),selectV.end());
  selectE.erase(selectE.begin(),selectE.end());
  selectF.erase(selectF.begin(),selectF.end());      
  RefineTotalV();
}

void
Extrude(){
  PreSub();
  ExtrudeCore();
  PostSub();
  Sync();
  PushOP(uGEOMETRY|uSELECTION);
}

void
ExtrudeCore(){

  int nowVertexSize=vertex.size();
  int nowSelectVSize=selectV.size();
  int edge1,edge2;
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    vertex.push_back(Vertex(0,0,0));
    subdividedVertex[1]++;
    SetChanged(1);
    vertex[vertex.size()-1].SetV(vertex[selectV[cnt1]]);    
  }
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    edge1=selectV[cnt1];
    for(int cnt2=0;cnt2<selectV.size();cnt2++){
      edge2=selectV[cnt2];
      DebugOut("edge1,edge2:" << edge1 << "," << edge2);
      if(FindEdge(edge1,edge2)!=-1){ // found 
	DebugOut("found edge");
	AddEdge(nowVertexSize+cnt1,nowVertexSize+cnt2);
      }
    }
  }
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    AddEdge(selectV[cnt1],nowVertexSize+cnt1);
  }
  selectV.erase(selectV.begin(),selectV.end());
  for(int cnt1=0;cnt1<nowSelectVSize;cnt1++){
    selectV.push_back(nowVertexSize+cnt1);
  }
  RefineTotalV();
}


void
PreSub(){
  nowLevel=subdivideLevel;
  for(int cnt1=0;cnt1<nowLevel;cnt1++){
    Subdivide(-1);
  }
}

void
PostSub(){
  for(int cnt1=0;cnt1<nowLevel;cnt1++){
    Subdivide(1);
  }
}

void
RecalcSub(int level){ // level is recalculating level now
  int faceS,faceE;
  int edgeS,edgeE;
  int vertexS,vertexE;

  Vertex tmpV;

  faceS=subdividedFace[level];
  faceE=subdividedFace[level+1];
  edgeS=subdividedEdge[level];
  edgeE=subdividedEdge[level+1];
  vertexS=subdividedVertex[level];
  vertexE=subdividedVertex[level+1];

  for(int cnt1=faceS;cnt1<faceE;cnt1++){
    tmpV=face[cnt1].CalcFacePoint();
    vertex[face[cnt1].facePointLink].SetV(tmpV);
  }
  for(int cnt1=edgeS;cnt1<edgeE;cnt1++){
    tmpV=edge[cnt1].CalcEdgePoint();
    vertex[edge[cnt1].edgePointLink].SetV(tmpV);
  }
  for(int cnt1=vertexS;cnt1<vertexE;cnt1++){
    if(vertex[cnt1].vertexPointLink!=-1){
      tmpV=vertex[cnt1].CalcVertexPoint();
      vertex[vertex[cnt1].vertexPointLink].SetV(tmpV);
    }
  }
}

void
FourToThree(){
  for(int cnt1=subdividedFace[0];cnt1<subdividedFace[1];cnt1++){
    if(face[cnt1].vertexLink.size()==4){
      AddEdge(face[cnt1].vertexLink[0],face[cnt1].vertexLink[2]);
      cnt1--;
    }
  }
}

void
SubdivideCore(int fStart,int fEnd,int eStart,int eEnd,int vStart,int vEnd){
  // Add Calculated Point
  
  // Face
  for(int cnt1=fStart;cnt1<fEnd;cnt1++){
    DebugOut ("Face[" << cnt1 << "]");
    vertex.push_back(Vertex(face[cnt1].CalcFacePoint()));	
    subdividedVertex[subdivideLevel+1]++;
    SetChanged(1);
    face[cnt1].facePointLink=vertex.size()-1;
  }
  
  // Edge
  for(int cnt2=eStart;cnt2<eEnd;cnt2++){
    /*
      #ifdef DEBUG
      cout << "Edge[" << cnt2 << "]" << endl;
      #endif
      */
    vertex.push_back(Vertex(edge[cnt2].CalcEdgePoint()));
    subdividedVertex[subdivideLevel+1]++;
    SetChanged(1);
    edge[cnt2].edgePointLink=vertex.size()-1;
    edge[cnt2].ConnectToFacePoint();
    //}
  }
  
  // Vertex
  for(int cnt3=vStart;cnt3<vEnd;cnt3++){
    /*
      #ifdef DEBUG
      cout << "Vertex[" << cnt3 << "]" << endl;
      #endif
      */
    if(vertex[cnt3].facePLink.size()>=1){
      vertex.push_back(Vertex(vertex[cnt3].CalcVertexPoint()));
      subdividedVertex[subdivideLevel+1]++;
      SetChanged(1);
      vertex[cnt3].vertexPointLink=vertex.size()-1;
      vertex[cnt3].ConnectToEdgePoint();
    }
  }      
}

void
DebugObjects(){
  cout << endl << "In DebugObjects" << endl;
  cout << "subFace " << endl;
  for(int cnt1=0;cnt1<subdividedFace.size();cnt1++){
    cout << subdividedFace[cnt1] << endl;
  }
  cout << "end" << endl;

  cout << "subEdge " << endl;
  for(int cnt1=0;cnt1<subdividedEdge.size();cnt1++){
    cout << subdividedEdge[cnt1] << endl;
  }
  cout << "end" << endl;

  cout << "subVertex " << endl;
  for(int cnt1=0;cnt1<subdividedVertex.size();cnt1++){
    cout << subdividedVertex[cnt1] << endl;
  }
  cout << "end" << endl;

  for(int cnt1=0;cnt1<face.size();cnt1++){
    cout << "face[" << cnt1 << "]" << endl;
    face[cnt1].Info();
  }
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    cout << "edge[" << cnt1 << "]" << endl;
    edge[cnt1].Info();
  }
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    cout << "vertex[" << cnt1 << "]" << endl;
    vertex[cnt1].Info();
  }
}

// levelTrans==+1 --> higher subdivide level display mode
// levelTrans==-1 --> lower subdivide level display mode
// levelTrans== 0 --> once subdivide
void
Subdivide(int levelTrans){  

  if(face.size()==0){
    DebugOut("Face not found.");
    return;
  }

  time_t startTime;
  time_t endTime;
  double workTime;
  
  int faceStart,faceEnd;
  int edgeStart,edgeEnd;
  int vertexStart,vertexEnd;

  // Time 
  startTime=time(NULL);

  switch(levelTrans){   
    
  case -1: // lower subdivide level ////////////////////////////
    if(subdivideLevel==0){
      DebugOut("now subdivide level is 0.");
      return;
    }

    // Face
    face.erase(&face[subdividedFace[subdivideLevel]],
	       &face[subdividedFace[subdivideLevel+1]]);
    //subdividedFace.pop_back();
    subdividedFace.erase(subdividedFace.end()-1);

    // Edge
    edge.erase(&edge[subdividedEdge[subdivideLevel]],
	       &edge[subdividedEdge[subdivideLevel+1]]);
    //subdividedEdge.pop_back();
    subdividedEdge.erase(subdividedEdge.end()-1);

    // Vertex
    vertex.erase(&vertex[subdividedVertex[subdivideLevel]],
		 &vertex[subdividedVertex[subdivideLevel+1]]);
    //subdividedVertex.pop_back();
    subdividedVertex.erase(subdividedVertex.end()-1);

    // change {face,edge,vertex}PointLink -> -1
    for(int cnt1=subdividedVertex[subdivideLevel-1];
	cnt1<subdividedVertex[subdivideLevel];cnt1++){
      vertex[cnt1].vertexPointLink=-1;
    }
    for(int cnt1=subdividedEdge[subdivideLevel-1];
	cnt1<subdividedEdge[subdivideLevel];cnt1++){
      edge[cnt1].edgePointLink=-1;
    }
    for(int cnt1=subdividedFace[subdivideLevel-1];
	cnt1<subdividedFace[subdivideLevel];cnt1++){
      face[cnt1].facePointLink=-1;
    }
    
    subdivideLevel-=1;
    DebugOut("now subdivide level is " << subdivideLevel);
    break;
    
  case +1: // higher level subdivide level /////////////////////////

    nowStatus=sSUBDIVIDE;

    // important !!!!!!!!
    // preallocation preserve validating iterator !! 
    vertex.reserve(2*vertex.size()+edge.size()+face.size());
    edge.reserve(edge.size()+2*edge.size()+face.size()*4);
    face.reserve(face.size()*4);
    
    // OK
    faceStart=subdividedFace[subdivideLevel];
    faceEnd=subdividedFace[subdivideLevel+1];
    edgeStart=subdividedEdge[subdivideLevel];
    edgeEnd=subdividedEdge[subdivideLevel+1];
    vertexStart=subdividedVertex[subdivideLevel];
    vertexEnd=subdividedVertex[subdivideLevel+1];

    subdividedVertex.push_back(vertexEnd);
    subdividedEdge.push_back(edgeEnd);
    subdividedFace.push_back(faceEnd);

    subdivideLevel+=1;    
    SubdivideCore(faceStart,faceEnd,edgeStart,edgeEnd,vertexStart,vertexEnd);

    DebugOut("now subdivide level is " << subdivideLevel);
    nowStatus=sNORMAL;
    break;
    
  case 0: // once subdivide  /////////////////////////////////////

    selectF.erase(selectF.begin(),selectF.end());  
    selectE.erase(selectE.begin(),selectE.end());
    selectV.erase(selectV.begin(),selectV.end());
    RefineTotalV();

    if(subdivideLevel==0){ // need subdivide
      nowStatus=sSUBDIVIDE;
      Subdivide(1);
      Subdivide(0);

      DebugOut("now subdivide level is " << subdivideLevel);
      nowStatus=sNORMAL;

    }else { // subdivideLevel>0 // need no subdivide 
      face.erase(&(face[0]),&(face[subdividedFace[1]]));
      edge.erase(&(edge[0]),&(edge[subdividedEdge[1]]));
      vertex.erase(&(vertex[0]),&(vertex[subdividedVertex[1]]));      

      // fix links in objects
      for(int cnt1=0;cnt1<face.size();cnt1++){
	for(int cnt2=0;cnt2<face[cnt1].edgeLink.size();cnt2++){
	  face[cnt1].edgeLink[cnt2]-=subdividedEdge[1];
	}
	for(int cnt2=0;cnt2<face[cnt1].vertexLink.size();cnt2++){
	  face[cnt1].vertexLink[cnt2]-=subdividedVertex[1];
	}
	if(face[cnt1].facePointLink!=-1){
	  face[cnt1].facePointLink-=subdividedVertex[1];
	}
      }


      for(int cnt1=0;cnt1<edge.size();cnt1++){
	for(int cnt2=0;cnt2<edge[cnt1].vertexLink.size();cnt2++){
	  edge[cnt1].vertexLink[cnt2]-=subdividedVertex[1];
	}
	for(int cnt2=0;cnt2<edge[cnt1].facePLink.size();cnt2++){
	  edge[cnt1].facePLink[cnt2]-=subdividedFace[1];
	}
	if(edge[cnt1].edgePointLink!=-1){
	  edge[cnt1].edgePointLink-=subdividedVertex[1];
	}
      }

      for(int cnt1=0;cnt1<vertex.size();cnt1++){
	for(int cnt2=0;cnt2<vertex[cnt1].edgePLink.size();cnt2++){
	  vertex[cnt1].edgePLink[cnt2]-=subdividedEdge[1];
	}
	for(int cnt2=0;cnt2<vertex[cnt1].facePLink.size();cnt2++){
	  vertex[cnt1].facePLink[cnt2]-=subdividedFace[1];
	}
	if(vertex[cnt1].vertexPointLink!=-1){
	  vertex[cnt1].vertexPointLink-=subdividedVertex[1];
	}
      }      

      // fix objects number in sudividedNumber{V,E,F}
      // dont need cnt1=0 
      for(int cnt1=subdivideLevel+1;cnt1>0;cnt1--){
	subdividedFace[cnt1]-=subdividedFace[1];
	subdividedEdge[cnt1]-=subdividedEdge[1];
	subdividedVertex[cnt1]-=subdividedVertex[1];
      }

      subdividedFace.erase(subdividedFace.begin());
      subdividedEdge.erase(subdividedEdge.begin());
      subdividedVertex.erase(subdividedVertex.begin());
      
      subdivideLevel-=1;
      
      DebugOut("now subdivide level is " << subdivideLevel);
    } 
    break;
  }
  
  endTime=time(NULL);
  workTime=difftime(endTime,startTime);  
  DebugOut("Subdivide time : " << workTime);
}

void
SelectedSharp(){
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    vertex[selectV[cnt1]].MoreSharp();
  }
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    edge[selectE[cnt1]].MoreSharp();
  }
}

void
SelectedSmooth(){
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    vertex[selectV[cnt1]].MoreSmooth();
  }
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    edge[selectE[cnt1]].MoreSmooth();
  }
}

void
WeightPropagate(){
  for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
    for(int cnt2=0;cnt2<subdividedEdge[cnt1];cnt2++){
      vertex[edge[cnt2].edgePointLink].weight=
	(vertex[edge[cnt2].vertexLink[Left]].weight+
	 vertex[edge[cnt2].vertexLink[Right]].weight)/2;
    }
  }
}


// Print All Edge
void
PrintEdge(){
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    edge[cnt1].Print();
  }
  cout << "edge num:" << edge.size() << endl;
}

// Print All Vertex
void
PrintVertex(){
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    vertex[cnt1].Print();
  }
  cout << "vertex num:" << vertex.size() << endl;
}

// Print All Face
void
PrintFace(){
  for(int cnt1=0;cnt1<face.size();cnt1++){
    face[cnt1].Print();
  }
  cout << "face num:" << face.size() << endl;
}


void
PrintVec(vector<int> vec){
  cout << "print vec" << endl;
  for(int cnt1=0;cnt1<vec.size();cnt1++){
    cout << vec[cnt1] << endl;
  }
}

int
CheckSave(void) {
  if (!changed) return 1;

  int choice;
  choice=fl_choice("Current file has not been saved.\n"
		   "Would you like to save it now?",
		   "Cancel","Yes","No");
  switch(choice){
  case 0:
    return (0);
    break;
  case 1:
    SaveCB();
    return (1);
    break;
  case 2:
    return (1);
    break;
  }
}

void
PYGInit(){
  // initialize
  minX=-0.1;minY=-0.1;minZ=-0.1;
  maxX=0.1;maxY=0.1;maxZ=0.1;
  selectV.erase(selectV.begin(),selectV.end());
  selectE.erase(selectE.begin(),selectE.end());
  selectF.erase(selectF.begin(),selectF.end());
  RefineTotalV();
  strcpy(tmpExpFilename,"tmppygzz.pyg");
  // stack
  stackOP.erase(stackOP.begin(),stackOP.end());
  undoVertex.erase(undoVertex.begin(),undoVertex.end());
  undoEdge.erase(undoEdge.begin(),undoEdge.end());
  undoFace.erase(undoFace.begin(),undoFace.end());
  undoSelectV.erase(undoSelectV.begin(),undoSelectV.end());
  undoSelectE.erase(undoSelectE.begin(),undoSelectE.end());
  undoSelectF.erase(undoSelectF.begin(),undoSelectF.end());
  lastGEOOP=-1;
  lastSELOP=-1;
  lastCAMOP=-1;
  lastTEXOP=-1;
  lastOPIndex=-1;
  PushOP(uGEOMETRY|uSELECTION);

  //cameraList.erase(cameraList.begin(),cameraList.end());
  int num=cameraList.size();
  for(int cnt1=0;cnt1<num;cnt1++){
    DeleteView(0); // num-cnt1-1 also will be good
  }

  lightList.erase(lightList.begin(),lightList.end());
  vertex.erase(vertex.begin(),vertex.end());
  edge.erase(edge.begin(),edge.end());
  face.erase(face.begin(),face.end());
  subdividedVertex.erase(subdividedVertex.begin(),subdividedVertex.end());
  subdividedEdge.erase(subdividedEdge.begin(),subdividedEdge.end());
  subdividedFace.erase(subdividedFace.begin(),subdividedFace.end());
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);
  subdivideLevel=0;
  selectV.erase(selectV.begin(),selectV.end());
  selectE.erase(selectE.begin(),selectE.end());
  selectF.erase(selectF.begin(),selectF.end());
  RefineTotalV();
  //removeFace.erase(removeFace.begin(),removeFace.end());
}

void
NewCB(void) {
  if (changed)
    if (!CheckSave()) return;

  NewFile fileDialog;
  fileDialog.mainWin->set_modal();
  fileDialog.show();
  for (;;) {
    Fl::wait();
    Fl_Widget *o;
    while ((o = Fl::readqueue())) {
      if (o == fileDialog.ok) {
	filename[0] = '\0';
	SetChanged(1);
	SetChanged(0);
	PYGInit();
	fileDialog.CreateAllView();
	fileDialog.mainWin->hide();
	AlignViews();
	Fl::redraw();
	return;
      } else if (o == fileDialog.cancel || o == fileDialog.mainWin) {
	fileDialog.mainWin->hide();
	return;
      }
    }
  }
}

void
OpenCB(void) {
  char *newfile;

  if (changed)
    if (!CheckSave()) return;

#ifdef DEBUG
  cout << "In OpenCB;filename==" << filename << endl;
#endif
  newfile = fl_file_chooser("Open File?", "*.pyg", filename);
  if (newfile != NULL) Load(newfile,OPEN);
}

void
ChangedCB(void) {
  SetChanged(1);
}

void
SetChanged(int c) {
  if (c != changed) {
    char title[1024];
    char *slash;

    changed = c;

    strcpy(title,"Pygmalion:");
    if (filename[0] == '\0') strcat(title, "Untitled.pyg");
    else {
      slash = strrchr(filename, '/');
      if (slash == NULL) slash = strrchr(filename, '\\');

      if (slash != NULL) strcat(title, slash + 1);
      else strcat(title, filename);
    }
    if (changed) strcat(title, " (modified)");
    mainM->mainWin->label(title);
  }
}

void
SaveCB(void) {
  if (filename[0] == '\0') {
    // No filename - get one!
    SaveAsCB();
    return;
  } else {
#ifdef DEBUG
    cout << "In SaveCB;filename==" << filename << endl;
#endif
    Save(filename);
  }
}

void
SaveAsCB(void) {
  char *newfile;

  newfile = fl_file_chooser("Save File As?", "*.pyg", filename);
  if (newfile != NULL) Save(newfile);
}

void
QuitCB(void) {
  if (changed){
    if (!CheckSave()){
      return;
    }
  }
  
  mainM->mainWin->hide();  

  for(int cnt1=0;cnt1<glW.size();cnt1++){
    glW[cnt1]->hide();
    delete glW[cnt1];
    glW[cnt1]=NULL;
  }
  
  delete cameraSet;
  cameraSet=NULL;
  delete mainM;
  mainM=NULL;
  
  exit(0);
}

void
Load(char* name,int type){
  FILE* fp;
  char buf[8192];

  SetChanged(1);
  
  fp=fopen(name,"r");
  if(fp==NULL){
    fl_alert("Unable to open \'%s\' for reading!");
    //cout << "couldn't open." << endl;
    return;
  }

  fscanf(fp,"%s ",buf); // header
#ifdef DEBUG
  cout << "header checking..." << endl;
#endif
  if(strcmp("PYG",buf)){
    fl_alert("\'%s\' is not Pygmalion file!",name);
    return;
  }

  // needed offset by importing 
  vertexOffset=0;  
  edgeOffset=0;
  faceOffset=0;
  int tmpLevel;

  if(type==OPEN){ // Initialize
    strcpy(filename,name);
    PYGInit();
  }else { // type==Import
    vertexOffset=subdividedVertex[1];
    edgeOffset=subdividedEdge[1];
    faceOffset=subdividedFace[1];
    // Subdivide Level -> 0
    tmpLevel=subdivideLevel;
    for(int cnt1=0;cnt1<tmpLevel;cnt1++){
      Subdivide(-1);
    }
  }

  // version,comment
  fscanf(fp,"%s ",buf); // version
#ifdef DEBUG
  cout << "version:" << buf << endl;
#endif
  //fscanf(fp,"%s ",buf); // comment
  fgets(buf,254,fp); // comment
#ifdef DEBUG
  cout << "comment==" << buf << endl;
#endif

  int vertexSize,edgeSize,faceSize,cameraSize,lightSize;

  fscanf(fp,"%d %d %d %d %d",&vertexSize,&edgeSize,&faceSize,
	 &cameraSize,&lightSize);
#ifdef DEBUG
  cout << "vSize,eSize,fSize,cSize,lSize" << vertexSize << ","
       << edgeSize << "," << faceSize << "," << cameraSize << "," 
       << lightSize << endl;
#endif
  vertex.reserve(vertexSize+vertexOffset);
  edge.reserve(edgeSize+edgeOffset);
  face.reserve(faceSize+faceOffset);
  
  // camera
  for(int cnt1=0;cnt1<cameraSize;cnt1++){
    cameraList.push_back(Camera());
    cameraList.back().ReadFile(fp);
    CreateGLWUI();
  }
  // light
  for(int cnt1=0;cnt1<lightSize;cnt1++){
    lightList.push_back(Light());
    lightList.back().ReadFile(fp);    
  }
  // vertex
  for(int cnt1=0;cnt1<vertexSize;cnt1++){
    vertex.push_back(Vertex());
    vertex.back().ReadFile(fp);
  }
  // edge
  for(int cnt1=0;cnt1<edgeSize;cnt1++){
    edge.push_back(Edge());
    edge.back().ReadFile(fp);
  }
  // face
  for(int cnt1=0;cnt1<faceSize;cnt1++){
    face.push_back(Face());
    face.back().ReadFile(fp);
  }
  fclose(fp);
  SetChanged(0);

  subdividedVertex[1]=vertex.size();
  subdividedEdge[1]=edge.size();
  subdividedFace[1]=face.size();

  if(type==IMPORT){
    for(int cnt1=0;cnt1<tmpLevel;cnt1++){
      Subdivide(1);
    }
    for(int cnt1=vertexOffset;cnt1<subdividedVertex[1];cnt1++){
      selectV.push_back(cnt1);
    }
    RefineTotalV();
  }
  
  for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
    CheckBB(vertex[cnt1]);
  }
  
  //AlignViews();
  Fl::redraw();
  PushOP(uGEOMETRY|uSELECTION);
#ifdef DEBUG
  cout << "all done." << endl;
#endif
}

void
Save(char* name){
  FILE* fp;
  fp=fopen(name,"w");
  if(fp==NULL){
    fl_alert("Unable to create \'%s\' for writing!");
    //cout << "couldn't open." << endl;
    return;
  }
  strcpy(filename,name);

  fprintf(fp,"PYG\n");
  fprintf(fp,"version1.0\n"); // version1.0
  fprintf(fp,"# comment line \n");
  fprintf(fp,"%d %d %d %d %d\n",subdividedVertex[1],subdividedEdge[1],
	  subdividedFace[1],(int)cameraList.size(),(int)lightList.size());
  // camera
  for(int cnt1=0;cnt1<cameraList.size();cnt1++){
    cameraList[cnt1].WriteFile(fp);
  }
  // light
  for(int cnt1=0;cnt1<lightList.size();cnt1++){
    lightList[cnt1].WriteFile(fp);
  }
  // vertex
  for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
    vertex[cnt1].WriteFile(fp);
  }
  // edge
  for(int cnt1=0;cnt1<subdividedEdge[1];cnt1++){
    edge[cnt1].WriteFile(fp);
  }
  // face
  for(int cnt1=0;cnt1<subdividedFace[1];cnt1++){
    face[cnt1].WriteFile(fp);
  }
  fclose(fp);
  changed=1;
  SetChanged(0);
}

void
CancelCB(){
  return;
}

void
ExportPOVCore(int level,int povFlag){
  char tmp[255];
  int cnt1;

  time_t startTime;
  time_t endTime;
  double workTime;
  // Time 
  startTime=time(NULL);

  DebugOut("In ExportPOVCore;filename==" << filename);
  if(!strcmp(filename,"")){
    sprintf(expFilename,"Untitled.pov");
  }else{    
    DebugOut("strlen==" << strlen(filename));
    for(cnt1=0;cnt1<strlen(filename)-4;cnt1++){
      tmp[cnt1]=filename[cnt1];
    }
    tmp[cnt1]='\0';
    DebugOut("tmp==" << tmp);
    sprintf(expFilename,"%s.pov",tmp);
  }
  if(povFlag==TEMP_EXPORT){
    sprintf(expFilename,"%s",tmpExpFilename);
  }
#ifdef DEBUG
  cout << "Export POV-Ray core" << endl;
  cout << "povFile==" << expFilename << endl;
#endif

  

  FILE* fp;
  int flag;
  fp=fopen(expFilename,"r"); 
  if(fp!=NULL && povFlag!=TEMP_EXPORT){
    char mes[255];
    sprintf(mes,"%s already exist. Overwrite?",expFilename);
    flag=fl_choice(mes,"yes","no",NULL);
    if(flag==1){ // no
      DebugOut("In ExportPOV;return");
      return;
    }else { // 2==yes
      fclose(fp);
      fp=fopen(expFilename,"w");
      DebugOut("Willing write...");
    }
  }else{
    fp=fopen(expFilename,"w");
    DebugOut("Willing write also...");
  }
  
  int dif=level-subdivideLevel;
  for(cnt1=0;cnt1<dif;cnt1++){
    Subdivide(1);
  }


  fprintf(fp,"// Generated by Pygmalion\n\n"); 
  fprintf(fp, "camera {\n");
  if(cameraList[outputC].type==PERSPECTIVE){
    fprintf(fp,"\tperspective\n");
    fprintf(fp,"\tlocation <%f, %f, %f>\n",cameraList[outputC].camera.X(),
	    cameraList[outputC].camera.Y(),-cameraList[outputC].camera.Z());
    fprintf(fp,"\tright %d*x\n",cameraList[outputC].width);
    fprintf(fp,"\tup %d*y\n",cameraList[outputC].height);
    fprintf(fp,"\tlook_at <%f, %f, %f>\n",cameraList[outputC].view.X(),
	    cameraList[outputC].view.Y(),-cameraList[outputC].view.Z());
    fprintf(fp,"\tangle %f\n",OGLAng2POVAng(cameraList[outputC],
					    cameraList[outputC].eyeAngle));
  }else{ // ORTHO
    float cameraCoord1=cameraList[outputC].camera.X();
    float cameraCoord2=cameraList[outputC].camera.Y();
    float cameraCoord3=-cameraList[outputC].camera.Z(); // left
    int signX,signY,signZ;
    switch(cameraList[outputC].type){
    case XYMP:
      signX=0;signY=0;signZ=1;
      cameraCoord3=-maxZ; // left coordinate
      break;
    case XYPM:
      signX=0;signY=0;signZ=-1;
      cameraCoord3=-minZ; // left 
      break;
    case YZMP:
      signX=+1;signY=0;signZ=0;
      cameraCoord1=minX;
      break;
    case YZPM:
      signX=-1;signY=0;signZ=0;
      cameraCoord1=maxX;
      break;
    case XZMP:
      signX=0;signY=+1;signZ=0;
      cameraCoord2=minY;
      break;
    case XZPM:
      signX=0;signY=-1;signZ=0;
      cameraCoord2=maxY;
      break;
    }
    fprintf(fp,"\torthographic\n");
    fprintf(fp, "\tlocation <%f, %f, %f>\n",
	    cameraCoord1,cameraCoord2,cameraCoord3);
    if(cameraList[outputC].width>cameraList[outputC].height){
      fprintf(fp,"\tright <2.0*%d/%d/%f, 0, 0>\n",
	      cameraList[outputC].width,
	      cameraList[outputC].height,
	      cameraList[outputC].zoom);
      fprintf(fp,"\tup <0, 2.0/%f, 0>\n",cameraList[outputC].zoom);
    }else{
      fprintf(fp,"\tright <2.0/%f, 0, 0>\n",cameraList[outputC].zoom);
      fprintf(fp,"\tup <0, 2.0*%d/%d/%f, 0>\n",
	      cameraList[outputC].height,
	      cameraList[outputC].width,
	      cameraList[outputC].zoom);
    }
    fprintf(fp,"\tdirection <0, 0, 1>\n");
    fprintf(fp,"\tlook_at <%f, %f, %f>\n",
	    cameraCoord1+signX,cameraCoord2+signY,cameraCoord3+signZ);
  }
  fprintf(fp, "}\n\n");

  Vertex* nowVertexP;
  Vector3d n0,n1,n2,n3; // vertex normals of a face
  // Face
  switch(povFlag){
  case TRIANGLE: // TRIANGLE
    fprintf(fp, "light_source { < 0, 1000, -2000 > color rgb <1, 1, 1> }\n");
    fprintf(fp, "light_source { < 2000, 100, 0 > color rgb <1, 1, 1> }\n");
    fprintf(fp, "light_source { < -500, -400, 2000 > color rgb <1, 1, 1> }\n\n");    
    fprintf(fp,"\n\n");
    fprintf(fp,"mesh {\n");

    for(cnt1=subdividedFace[level];cnt1<subdividedFace[level+1];cnt1++){
      if(face[cnt1].vertexLink.size()==3){
	fprintf(fp,"\ttriangle {\n"); 
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[0]].X(),
		vertex[face[cnt1].vertexLink[0]].Y(),
		-vertex[face[cnt1].vertexLink[0]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[1]].X(),
		vertex[face[cnt1].vertexLink[1]].Y(),
		-vertex[face[cnt1].vertexLink[1]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>}\n",
		vertex[face[cnt1].vertexLink[2]].X(),
		vertex[face[cnt1].vertexLink[2]].Y(),
		-vertex[face[cnt1].vertexLink[2]].Z());
      }else{
	fprintf(fp,"\ttriangle {\n"); 
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[0]].X(),
		vertex[face[cnt1].vertexLink[0]].Y(),
		-vertex[face[cnt1].vertexLink[0]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[1]].X(),
		vertex[face[cnt1].vertexLink[1]].Y(),
		-vertex[face[cnt1].vertexLink[1]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>}\n",
		vertex[face[cnt1].vertexLink[2]].X(),
		vertex[face[cnt1].vertexLink[2]].Y(),
		-vertex[face[cnt1].vertexLink[2]].Z());
	fprintf(fp,"\ttriangle {\n"); 
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[2]].X(),
		vertex[face[cnt1].vertexLink[2]].Y(),
		-vertex[face[cnt1].vertexLink[2]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>,\n",
		vertex[face[cnt1].vertexLink[3]].X(),
		vertex[face[cnt1].vertexLink[3]].Y(),
		-vertex[face[cnt1].vertexLink[3]].Z());
	fprintf(fp,"\t\t<%f, %f, %f>}\n",
		vertex[face[cnt1].vertexLink[0]].X(),
		vertex[face[cnt1].vertexLink[0]].Y(),
		-vertex[face[cnt1].vertexLink[0]].Z());
      }
    }
    break;

  case SMOOTH_TRIANGLE:    
  case TEMP_EXPORT:
    //vector<int> flag[vertex.size()]; // calculation flag

    fprintf(fp, "light_source { < %f, %f, %f > color rgb <1, 1, 1> }\n",
	    (cameraList[outputC].camera-cameraList[outputC].view).X()*1000
	    +cameraList[outputC].camera.X(),
	    (cameraList[outputC].camera-cameraList[outputC].view).Y()*1000
	    +cameraList[outputC].camera.Y(),
	    -(cameraList[outputC].camera-cameraList[outputC].view).Z()*1000
	    -cameraList[outputC].camera.Z());
    fprintf(fp,"\nmesh {\n");

#ifdef DEBUG
    cout << "smooth_triangle" << endl;
#endif
    for(int nowFace=subdividedFace[level];nowFace<subdividedFace[level+1];
	nowFace++){// start face
      if(face[nowFace].vertexLink.size()==4){ // case: 4 vertices
#ifdef DEBUG
	cout << "4 vertices face" << endl;
	cout << "nowFace==" << nowFace << endl;
#endif

	for(int cntTmp=0;cntTmp<6;cntTmp++){ //start vertex
	  int cnt2=cntTmp;
	  if(cnt2>2){ // for second triangle
	    cnt2-=1;
	  }// cnt2 mean triangle vertex
	  int nowVertex=face[nowFace].vertexLink[cnt2%4];
#ifdef DEBUG
	  cout << "nowVertex==" << nowVertex << endl;
#endif
	  nowVertexP=&vertex[nowVertex];
	  vector<int> startEdge; // start edge number
	  vector<int> checkCnt; 

	  // find start 2 edges
	  int tmpNowV,opV,nowEdge;
	  for(int cnt3=0;cnt3<face[nowFace].vertexLink.size();cnt3++){
	    tmpNowV=face[nowFace].vertexLink[cnt3];
	    opV=face[nowFace].vertexLink[(cnt3+1)%4]; // opposite vertex
	    nowEdge=face[nowFace].FindEdge(tmpNowV,opV);
	    if(!vertex[nowVertex].HaveEdge(nowEdge)){
	      continue;
	    }
	    startEdge.push_back(nowEdge);
	    checkCnt.push_back(cnt3);
	  }
	  if(checkCnt[0]==0 && checkCnt[1]==3){
#ifdef DEBUG
	    cout << "swap startEdge" << endl;
#endif
	    swap(startEdge[0],startEdge[1]);
	  }
#ifdef DEBUG
	  cout << "startEdge.size()==" << startEdge.size() << endl;
	  for(int tmp=0;tmp<startEdge.size();tmp++){
	    cout << "startEdge[" << tmp << "]==" << startEdge[tmp] << endl;
	  }
#endif
	  Vector3d totalNormal;
	  int faceSum=0;
	  Vector3d vec[3];
	  Vector3d tmpN;
	  vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					 AnotherVertex(nowVertex)];
	  vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					 AnotherVertex(nowVertex)];
	  
	  // can optimise.HasBreak...nowVertex-->cntTmp
	  //if(face[nowFace].HasBreak(nowVertex)){ 
	  if(cntTmp!=1 && cntTmp!=3){ // Done by Hidekuni Annaka 1/4/2000
	    vec[2]=vertex[nowVertex]
	      -vertex[face[nowFace].OppositeVertex(nowVertex)];
	    Vector3d tmpVec;
	    tmpVec.Cross(vec[0],vec[2]);
	    tmpVec.Normalize();
#ifdef DEBUG
	    tmpVec.Print();
#endif
	    if(isnan(tmpVec.coord[0])){
#ifdef DEBUG
	      cout << "prevent nan : set zero" << endl;
#endif
	      tmpVec.Set(0,0,0); // prevent nan
	    }
	    tmpN.Cross(vec[2],vec[1]);
	    tmpN.Normalize();
#ifdef DEBUG
	    tmpN.Print();
#endif
	    if(isnan(tmpN.coord[0])){
#ifdef DEBUG
	      cout << "prevent nan : set zero2" << endl;
#endif
	      tmpN.Set(0,0,0); // prevent nan
	    }
	    tmpN.Plus(tmpVec);
	    tmpN.Normalize();
	  }else{
	    tmpN.Cross(vec[0],vec[1]);
	    tmpN.Normalize(); // finish this face normal
	  }
	  if(isnan(tmpN.coord[0])){
#ifdef DEBUG
	    cout << "prevent nan:abort this face2 -> set dummy" << endl;
#endif
	    tmpN.Set(0,0,0);
	  }
	  
#ifdef DEBUG
	  cout << "base normal==";
	  tmpN.Print();
#endif
	  totalNormal.Plus(tmpN); // add this face's normal
	  faceSum++;
	  for(int cnt3=0;cnt3<startEdge.size();cnt3++){
#ifdef DEBUG
	    cout << "searching next face from edge[" << startEdge[cnt3] << "]..."<< endl;
#endif
	    int oldEdge=startEdge[cnt3],newEdge;
	    int oldFace=nowFace,newFace;	    
	   
	    vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					   AnotherVertex(nowVertex)];
	    vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					   AnotherVertex(nowVertex)];
	    if(cnt3!=0){
	      swap(vec[0],vec[1]);
	    }
	    int whileFlag=1;
	    while(edge[oldEdge].facePLink.size()==2 && whileFlag){
#ifdef DEBUG
	      cout << "found new face." << endl;
#endif
	      newFace=edge[oldEdge].AnotherFace(oldFace); // newFace
#ifdef DEBUG
	      cout << "oldEdge,oldFace,newFace==" << oldEdge << ", "
		   << oldFace << ", " << newFace << endl;
#endif
	      if(newFace==nowFace){ // prevent loop
#ifdef DEBUG
		cout << "but newFace is nowFace. so end nowFace." << endl;
#endif
		cnt3=startEdge.size();
		whileFlag=0;
		continue;
	      }
	      vec[1]=vec[0];

	      int tmpEdge;
	      for(int cnt4=0;cnt4<nowVertexP->edgePLink.size();cnt4++){
		tmpEdge=nowVertexP->edgePLink[cnt4];
		if(face[newFace].HaveEdge(tmpEdge) && tmpEdge!=oldEdge){
		  newEdge=tmpEdge; // newEdge
		  cnt4=nowVertexP->edgePLink.size(); // exit this for-loop
		}
	      }
#ifdef DEBUG
	      cout << "newEdge==" << newEdge << endl;
#endif
	      vec[0]=vertex[nowVertex]
		-vertex[edge[newEdge].AnotherVertex(nowVertex)]; // another vector
	      if(face[newFace].HasBreak(nowVertex)){
#ifdef DEBUG
		cout << "nowVertex in newFace has break" << endl;
#endif
		vec[2]=vertex[nowVertex]
		  -vertex[face[newFace].OppositeVertex(nowVertex)];
		Vector3d tmpVec;
		tmpVec.Cross(vec[0],vec[2]);
		tmpVec.Normalize();
		if(isnan(tmpVec.coord[0])){
#ifdef DEBUG
		  cout << "prevent nan : through this face" << endl;
#endif
		  tmpVec.Set(0,0,0); // prevent nan
		}
		tmpN.Cross(vec[2],vec[1]);
		tmpN.Normalize();
		if(isnan(tmpN.coord[0])){
#ifdef DEBUG
		  cout << "prevent nan : through this face2" << endl;
#endif
		  tmpN.Set(0,0,0); // prevent nan
		}
		tmpN.Plus(tmpVec);
		tmpN.Normalize();
	      }else{
		tmpN.Cross(vec[0],vec[1]);
		tmpN.Normalize();
	      }
	      if(isnan(tmpN.coord[0])){
#ifdef DEBUG
		cout << "prevent NAN : set zero" << endl;
#endif
		tmpN.Set(0,0,0);
		faceSum--;
	      }
#ifdef DEBUG
	      cout << "tmpN==";
	      tmpN.Print();
#endif
	      if(cnt3==0){
		totalNormal.Plus(tmpN); // add this face's normal
	      }else{
		totalNormal.Minus(tmpN); // add this face's normal
	      }
	      faceSum++;
#ifdef DEBUG
	      cout << "now totalNormal==";
	      totalNormal.Print();
#endif
	      oldFace=newFace; // store face number
	      oldEdge=newEdge; // store edge number
	    }
#ifdef DEBUG
	    cout << "end search." << endl;
#endif
	  }
	  totalNormal.Normalize();
	  if(isnan(totalNormal.coord[0])){
#ifdef DEBUG
	    cout << "set totalNormal 0,0,0" << endl;
#endif
	    totalNormal.Set(0,0,0);
	  }
#ifdef DEBUG
	  printf("\tfaceSum==%d\n",faceSum);
	  printf("\tAverage normal==");
	  totalNormal.Print();
#endif
	  // print POV smooth_triangle
	  if(cntTmp==0 || cntTmp==3){ // first vertex of triangle
	    fprintf(fp,"\tsmooth_triangle {\n");
	  }
	  fprintf(fp,"\t\t<%f, %f, %f>,<%f, %f, %f>",
		  vertex[nowVertex].X(),
		  vertex[nowVertex].Y(),
		  -vertex[nowVertex].Z(),
		  totalNormal.X(),
		  totalNormal.Y(),
		  -totalNormal.Z()); 	  
	  if(cntTmp==2 || cntTmp==5){ // first vertex of triangle
	    fprintf(fp,"}\n");
	  }else{
	    fprintf(fp,",\n");
	  }
	}

      }else{ // 3 vertices face
#ifdef DEBUG
	cout << "3 vertices face" << endl;
	cout << "nowFace==" << nowFace << endl;
#endif
	fprintf(fp,"\tsmooth_triangle {\n");
	for(int cnt2=0;cnt2<face[nowFace].vertexLink.size();
	    cnt2++){ //start vertex
	  int nowVertex=face[nowFace].vertexLink[cnt2];
#ifdef DEBUG
	  cout << "nowVertex==" << nowVertex << endl;
#endif	
	  nowVertexP=&vertex[nowVertex];
	  vector<int> startEdge; // start edge number

	  int flag=0; 
	  int nowEdge;
	  for(int cnt3=0;cnt3<face[nowFace].edgeLink.size();cnt3++){
	    nowEdge=face[nowFace].edgeLink[cnt3];
	    if(!vertex[nowVertex].HaveEdge(nowEdge)){
	      continue;
	    }
	    startEdge.push_back(nowEdge);
	    flag+=cnt3;
	  }	  
	  if(flag==2){ // edgeLink[0] and edgeLink[2]
	    swap(startEdge[0],startEdge[1]); // sort startEdge in a face
	  }
#ifdef DEBUG
	  cout << "startEdge.size()==" << startEdge.size() << endl;
	  for(int tmp=0;tmp<startEdge.size();tmp++){
	    cout << "startEdge[" << tmp << "]==" << startEdge[tmp] << endl;
	  }
#endif
	  Vector3d totalNormal;
	  int faceSum=0;
	  Vector3d vec[3];
	  Vector3d tmpN,tmpN2;
	  //Vector3d tmpVec;
	  vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					 AnotherVertex(nowVertex)];
	  vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					 AnotherVertex(nowVertex)];

	  tmpN2=vec[0].Cross(vec[1]);
	  tmpN2.Normalize();
#ifdef DEBUG
	  cout << "base normal==";
	  tmpN2.Print();
#endif
	  if(isnan(tmpN2.coord[0])){
#ifdef DEBUG
	    cout << "prevent nan : set zero" << endl;
#endif
	    tmpN.Set(0,0,0);
	  }
	  totalNormal.Plus(tmpN2); // add this face's normal
	  faceSum++;
	  for(int cnt3=0;cnt3<startEdge.size();cnt3++){
#ifdef DEBUG
	    cout << "searching next face from edge[" << startEdge[cnt3] << "]..."<< endl;
#endif
	    int oldEdge=startEdge[cnt3],newEdge;
	    int oldFace=nowFace,newFace;	    
	    vec[0]=vertex[nowVertex]-vertex[edge[startEdge[0]].
					   AnotherVertex(nowVertex)];
	    vec[1]=vertex[nowVertex]-vertex[edge[startEdge[1]].
					   AnotherVertex(nowVertex)];
	    if(cnt3!=0){
	      swap(vec[0],vec[1]);
	    }
	    int whileFlag=1;
	    while(edge[oldEdge].facePLink.size()==2 && whileFlag){
#ifdef DEBUG
	      cout << "found new face." << endl;
#endif
	      newFace=edge[oldEdge].AnotherFace(oldFace); // newFace
#ifdef DEBUG
	      cout << "oldEdge,oldFace,newFace==" << oldEdge << ", "
		   << oldFace << ", " << newFace << endl;
#endif
	      if(newFace==nowFace){ // prevent loop
#ifdef DEBUG
		cout << "but newFace is nowFace. so end nowFace." << endl;
#endif
		cnt3=startEdge.size();
		whileFlag=0;
		continue;
	      }
	      vec[1]=vec[0];
	      
	      int tmpEdge;
	      for(int cnt4=0;cnt4<nowVertexP->edgePLink.size();cnt4++){
		tmpEdge=nowVertexP->edgePLink[cnt4];
		if(face[newFace].HaveEdge(tmpEdge) && tmpEdge!=oldEdge){
		  newEdge=tmpEdge; // newEdge
		  cnt4=nowVertexP->edgePLink.size(); // exit this for-loop
		}
	      }
#ifdef DEBUG
	      cout << "newEdge==" << newEdge << endl;
#endif
	      vec[0]=vertex[nowVertex]
		-vertex[edge[newEdge].AnotherVertex(nowVertex)]; // another vector
	      if(face[newFace].HasBreak(nowVertex)){
#ifdef DEBUG
		cout << "nowVertex in newFace has break" << endl;
#endif
		vec[2]=vertex[nowVertex]
		  -vertex[face[newFace].OppositeVertex(nowVertex)];
		Vector3d tmpVec;
		tmpVec.Cross(vec[0],vec[2]);
		tmpVec.Normalize();
		if(isnan(tmpVec.coord[0])){
#ifdef DEBUG
		  cout << "prevent nan : through this face" << endl;
#endif
		  tmpVec.Set(0,0,0); // prevent nan
		}
		tmpN.Cross(vec[2],vec[1]);
		tmpN.Normalize();
		if(isnan(tmpN.coord[0])){
#ifdef DEBUG
		  cout << "prevent nan : set zero2" << endl;
#endif
		  tmpN.Set(0,0,0); // prevent nan
		}
		tmpN.Plus(tmpVec);
		tmpN.Normalize();
	      }else{
		tmpN.Cross(vec[0],vec[1]);
		tmpN.Normalize();
	      }
	      if(isnan(tmpN.coord[0])){
#ifdef DEBUG
		cout << "prevent nan:abort this face2 -> set dummy" << endl;
#endif
		tmpN.Set(0,0,0);
	      }
	      
	      if(cnt3==0){
		totalNormal.Plus(tmpN); // add this face's normal
	      }else{
		totalNormal.Minus(tmpN); // add this face's normal
	      }
	      faceSum++;
#ifdef DEBUG
	      cout << "now totalNormal==";
	      totalNormal.Print();
#endif
	      oldFace=newFace; // store face number
	      oldEdge=newEdge; // store edge number
	    }
#ifdef DEBUG
	    cout << "end search." << endl;
#endif
	  }
	  totalNormal.Normalize();
	  if(isnan(totalNormal.coord[0])){
#ifdef DEBUG
	    cout << "set totalNormal 0,0,0" << endl;
#endif
	    totalNormal.Set(0,0,0);
	  }
#ifdef DEBUG
	  printf("\tfaceSum==%d\n",faceSum);
	  printf("\tAverage normal==");
	  totalNormal.Print();
#endif
	  fprintf(fp,"\t\t<%f, %f, %f>,<%f, %f, %f>",
		  vertex[nowVertex].X(),
		  vertex[nowVertex].Y(),
		  -vertex[nowVertex].Z(),
		  totalNormal.X(),
		  totalNormal.Y(),
		  -totalNormal.Z()); 	  
	  if(cnt2!=2){
	    fprintf(fp,",\n");
	  }else{
	    fprintf(fp,"}\n");
	  }
	}

      }
    }
    break;
  }

  fprintf(fp,"\ttexture {\n"); // texture
  fprintf(fp,"\t\tpigment {\n"); // pigment
  fprintf(fp,"\t\t\t rgb <1,1,1>\n");
  fprintf(fp,"\t\t}\n"); // pigment
  fprintf(fp,"\t\tfinish {\n"); // finish
  fprintf(fp,"\t\t\tdiffuse 0.5\n");
  fprintf(fp,"\t\t\tambient 0.1\n");
  fprintf(fp,"\t\t\tspecular .5 roughness .05\n");
  fprintf(fp,"\t\t}\n"); // finish
  fprintf(fp,"\t}\n"); // texture
  fprintf(fp,"}\n"); // union

  fclose(fp);
  for(cnt1=0;cnt1<dif;cnt1++){
    Subdivide(-1);
#ifdef DEBUG
    cout << "unsub..." << endl;
#endif
  }

  endTime=time(NULL);
  workTime=difftime(endTime,startTime);  
#ifdef DEBUG
  cout << "Export time : " << workTime << endl;
  cout << "Exported." << endl;
#endif
}

int
GetString(char* exportType,char* defaultSub){
#ifdef DEBUG
  cout << defaultSub << endl;
#endif
  int mouseX=Fl::event_x_root();
  int mouseY=Fl::event_y_root();
  Fl_Window povExportW(mouseX,mouseY,225,65+30,exportType);
  Fl_Int_Input povInput(120,5,100,20,"Subdivision Level");
  povInput.value(defaultSub);
  Fl_Choice camChoice(120,28,90,30,"select camera");
  char camIndex[1024];
  sprintf(camIndex,"");
  for(int cnt1=0;cnt1<cameraList.size();cnt1++){
    strcat(camIndex,cameraList[cnt1].name);
    strcat(camIndex," - ");
    switch(cameraList[cnt1].type){
      //case NORMAL:
    case PERSPECTIVE:
      strcat(camIndex,"Perspective");
      break;
      //case XY:
    case XYMP:
      strcat(camIndex,"XY-Ortho:Z ->+");
      break;
    case XYPM:
      strcat(camIndex,"XY-Ortho:Z +>-");
      break;
      //case YZ:
    case YZMP:
      strcat(camIndex,"YZ-Ortho:X ->+");
      break;
    case YZPM:
      strcat(camIndex,"YZ-Ortho:X +>-");
      break;
      //case XZ:
    case XZMP:
      strcat(camIndex,"XZ-Ortho:Y ->+");
      break;      
    case XZPM:
      strcat(camIndex,"XZ-Ortho:Y +>-");
      break;      
    }
    strcat(camIndex," - ");
    char widthT[10];
    char heightT[10];
    sprintf(widthT,"%dx",cameraList[cnt1].width);
    sprintf(heightT,"%d",cameraList[cnt1].height);
    strcat(camIndex,widthT);
    strcat(camIndex,heightT);
    strcat(camIndex," - ");
    char camOrder[20];
    sprintf(camOrder,"camera[%d]",cnt1);
    strcat(camIndex,camOrder);
    strcat(camIndex,"|");
  }
  camChoice.add(camIndex);
  camChoice.value(0);
  Fl_Return_Button ok(30,30+30,70,30,"ok");
  Fl_Button cancel(120,30+30,70,30,"cancel");
  povExportW.hotspot(&cancel);
  //cancel.callback((Fl_Callback*)CancelCB);
  //ok.callback((Fl_Callback*)ExportPOVCore);
  
  povExportW.end();
  povExportW.set_modal();
  povExportW.show();
  for (;;) {
    Fl::wait();
    Fl_Widget *o;
    while ((o = Fl::readqueue())) {
      if (o == &ok) {
        strcpy(defaultSub,povInput.value());
	outputC=camChoice.value();
        return atoi(povInput.value());
      } else if (o == &cancel || o == &povExportW) {
        return -1;
      }
    }
  }
}

void
ExportOFF(){
  DebugOut("ExportOFF");
  char offFile[255];
  char tmp[255];
  int cnt1;

  time_t startTime;
  time_t endTime;
  double workTime;
  // Time 
  startTime=time(NULL);

  int level=GetString("OFF Export",defaultSub);
  if(level<0){
    return;
  }
  DebugOut("In ExportOFF;filename==" << filename);
  if(!strcmp(filename,"")){
    sprintf(offFile,"Untitled.off");
  }else{
    
    DebugOut("strlen==" << strlen(filename));
    for(cnt1=0;cnt1<strlen(filename)-4;cnt1++){
      tmp[cnt1]=filename[cnt1];
    }
    tmp[cnt1]='\0';
    DebugOut("tmp==" << tmp);
    //strncpy(tmp,filename,(int)strlen(filename)-4);
    sprintf(offFile,"%s.off",tmp);
  }

  FILE* fp;
  int flag;
  fp=fopen(offFile,"r"); 
  if(fp!=NULL){
    char mes[255];
    sprintf(mes,"%s already exist. Overwrite?",offFile);
    flag=fl_choice(mes,"yes","no",NULL);
    if(flag==1){ // no
#ifdef DEBUG
      cout << "In ExportOFF;return" << endl;
#endif
      return;
    }else { // 2==yes
      fclose(fp);
      fp=fopen(offFile,"w");
#ifdef DEBUG
      cout << "Willing write..." << endl;
#endif
    }
  }else{
    fp=fopen(offFile,"w");
#ifdef DEBUG
    cout << "Willing write also..." << endl;
#endif
  }
  
  int dif=level-subdivideLevel;
  for(cnt1=0;cnt1<dif;cnt1++){
    Subdivide(1);
  }
  
  fprintf(fp,"OFF\n");
  fprintf(fp,"%d %d %d\n",
	  subdividedVertex[level+1]-subdividedVertex[level],
	  subdividedFace[level+1]-subdividedFace[level],
	  subdividedEdge[level+1]-subdividedEdge[level]);
  for(int cnt1=subdividedVertex[level];cnt1<subdividedVertex[level+1];cnt1++){
    fprintf(fp,"%f %f %f\n",
	    vertex[cnt1].X(),
	    vertex[cnt1].Y(),
	    vertex[cnt1].Z());
  }
  for(int cnt1=subdividedFace[level];cnt1<subdividedFace[level+1];cnt1++){
    if(face[cnt1].vertexLink.size()==3){
      fprintf(fp,"%d %d %d %d\n",3,
	      face[cnt1].vertexLink[0]-subdividedVertex[level],
	      face[cnt1].vertexLink[1]-subdividedVertex[level],
	      face[cnt1].vertexLink[2]-subdividedVertex[level]);
    }else {
      fprintf(fp,"%d %d %d %d %d\n",4,
	      face[cnt1].vertexLink[0]-subdividedVertex[level],
	      face[cnt1].vertexLink[1]-subdividedVertex[level],
	      face[cnt1].vertexLink[2]-subdividedVertex[level],
	      face[cnt1].vertexLink[3]-subdividedVertex[level]);
    }
  }  
  fclose(fp);
}

void
ImportPYG(){
  DebugOut("Import PYG");
  char *newfile;  

  newfile = fl_file_chooser("Import File?", "*.pyg", filename);
  if (newfile != NULL) Load(newfile,IMPORT);
}

// triangle
void
ExportPOV(){
  DebugOut("POV-Ray triangle Export");

  int level=GetString("POV-Ray Export:triangle",defaultSub);
  if(level>=0){
    ExportPOVCore(level,TRIANGLE);
  }
  
  DebugOut(defaultSub);
}


void
ExportPOVSmooth(){
  DebugOut("POV-Ray smooth_triangle Export");


  int level=GetString("POV-Ray Export:smooth_triangle",defaultSub);
  if(level>=0){
    ExportPOVCore(level,SMOOTH_TRIANGLE);
  }
  DebugOut(defaultSub);
}


// Test Main Procedure
int
main(int argc,char** argv){
  // always subdividedNumber{V,E,F}[0]==0
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);

  // number of objects in subdivide level 0 
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);
  sprintf(defaultSub,"3"); 
  mainM = new MainMenu;

  //mainM->undoM->activate();
  //mainM->redoM->activate();
  undoMA=mainM->undoM;
  redoMA=mainM->redoM;
  restX=mainM->restrictXM;
  restY=mainM->restrictYM;
  restZ=mainM->restrictZM;

  mainM->mainWin->show();
  mainM->mainWin->position(0,0);

  cameraSet = new CameraSet; // camera property dialog
  userSetting = new UserSetting; // User Setting dialog

  PYGInit();

  // if given input file,
  if (argc > 1) Load(argv[1],OPEN);  

  Fl::run();

  return 0;
}
